from abc import abstractmethod
from typing import List

import pandas as pd  # type: ignore
from sklearn.model_selection import train_test_split  # type: ignore


class TrainDataset:
    def __init__(self, df: pd.DataFrame, features: List[str], target: str):
        self.df = df
        self.features = features
        self.target = target
        x = self.df.drop([target], axis=1)
        y = self.df[target]

        # Prepare train, test and validation datasets
        self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(x, y, test_size=0.30, random_state=42)
        self.x_test, self.x_val, self.y_test, self.y_val = train_test_split(
            self.x_test, self.y_test, test_size=0.5, random_state=42
        )


class AutoMLModel:
    # model types
    CLASSIFIER = "classifier"
    REGRESSOR = "regressor"

    def __init__(self) -> None:
        self.model = None
        self.score = 0
        self.feature_importances = None
        self.explainer = None

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Name of the AutoML model that will be shown to the user.
        :return: Name
        """

    @property
    @abstractmethod
    def model_type(self) -> str:
        """
        Defines the model type
        :return: Model type
        """

    @abstractmethod
    def train(self, dataset: TrainDataset) -> None:
        """
        Trains an AutoML model with the dataset provided

        :param dataset: Dataset to be used in training
        :return: None
        """

    def compare_score(self, score: float) -> bool:
        """
        Compares the model score to define the best model
        :param score: Score of another model to be compared
        :return: True/False is the score is better/worse
        """



from dataclasses import dataclass
from typing import Any, Dict, Optional

from dbt.adapters.bigquery.connections import (  # type:ignore
    BigQueryConnectionManager,
    BigQueryCredentials,
)


@dataclass
class LayerBigQueryCredentials(BigQueryCredentials):
    # Layer config json creds
    layer_configfile: Optional[str] = None
    layer_configfile_json: Optional[Dict[str, Any]] = None
    layer_project: Optional[str] = None
    layer_api_key: Optional[str] = None

    @property
    def type(self) -> str:
        return "layer_bigquery"


class LayerBigQueryConnectionManager(BigQueryConnectionManager):
    TYPE = "layer_bigquery"



from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier  # type: ignore
from sklearn.linear_model import LinearRegression, RidgeClassifier  # type: ignore
from sklearn.metrics import accuracy_score, r2_score  # type: ignore
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor  # type: ignore

from .base_model import AutoMLModel, TrainDataset


class ScikitLearnLinearRegression(AutoMLModel):

    name = "Scikit-Learn LinearRegression"
    model_type = AutoMLModel.REGRESSOR

    def __init__(self) -> None:
        super().__init__()

    def train(self, ds: TrainDataset) -> None:
        model = LinearRegression(normalize=True, fit_intercept=False, copy_X=True)
        model.fit(ds.x_train, ds.y_train)
        y_pred = model.predict(ds.x_test)
        model_accuracy = r2_score(ds.y_test, y_pred)

        self.score = model_accuracy
        self.model = model

    def compare_score(self, score: float) -> bool:
        return self.score > score


class ScikitLearnDecisionTreeRegressor(AutoMLModel):

    name = "Scikit-Learn DecisionTreeRegressor"
    model_type = AutoMLModel.REGRESSOR

    def __init__(self) -> None:
        super().__init__()

    def train(self, ds: TrainDataset) -> None:
        model = DecisionTreeRegressor(max_depth=7)
        model.fit(ds.x_train, ds.y_train)
        y_pred = model.predict(ds.x_test)
        model_accuracy = r2_score(ds.y_test, y_pred)

        self.score = model_accuracy
        self.model = model
        self.feature_importances = model.feature_importances_

    def compare_score(self, score: float) -> bool:
        return self.score > score


class ScikitLearnRandomForestClassifier(AutoMLModel):
    name = "Scikit-Learn RandomForestClassifier"
    model_type = AutoMLModel.CLASSIFIER

    def __init__(self) -> None:
        super().__init__()

    def train(self, ds: TrainDataset) -> None:
        model = RandomForestClassifier()
        model.fit(ds.x_train, ds.y_train)
        predictions = model.predict(ds.x_test)
        model_accuracy = accuracy_score(ds.y_test, predictions)

        self.score = model_accuracy
        self.model = model
        self.feature_importances = model.feature_importances_

    def compare_score(self, score: float) -> bool:
        return self.score > score


class ScikitLearnRidgeClassifier(AutoMLModel):
    name = "Scikit-Learn RidgeClassifier"
    model_type = AutoMLModel.CLASSIFIER

    def __init__(self) -> None:
        super().__init__()

    def train(self, ds: TrainDataset) -> None:
        model = RidgeClassifier()
        model.fit(ds.x_train, ds.y_train)
        predictions = model.predict(ds.x_test)
        model_accuracy = accuracy_score(ds.y_test, predictions)

        self.score = model_accuracy
        self.model = model

    def compare_score(self, score: float) -> bool:
        return self.score > score


class ScikitLearnDecisionTreeClassifier(AutoMLModel):
    name = "Scikit-Learn DecisionTreeClassifier"
    model_type = AutoMLModel.CLASSIFIER

    def __init__(self) -> None:
        super().__init__()

    def train(self, ds: TrainDataset) -> None:
        model = DecisionTreeClassifier(max_depth=5)
        model.fit(ds.x_train, ds.y_train)
        predictions = model.predict(ds.x_test)
        model_accuracy = accuracy_score(ds.y_test, predictions)

        self.score = model_accuracy
        self.model = model
        self.feature_importances = model.feature_importances_

    def compare_score(self, score: float) -> bool:
        return self.score > score


class ScikitLearnAdaBoostClassifier(AutoMLModel):
    name = "Scikit-Learn AdaBoostClassifier"
    model_type = AutoMLModel.CLASSIFIER

    def __init__(self) -> None:
        super().__init__()

    def train(self, ds: TrainDataset) -> None:
        model = AdaBoostClassifier()
        model.fit(ds.x_train, ds.y_train)
        predictions = model.predict(ds.x_test)
        model_accuracy = accuracy_score(ds.y_test, predictions)

        self.score = model_accuracy
        self.model = model
        self.feature_importances = model.feature_importances_

    def compare_score(self, score: float) -> bool:
        return self.score > score
